﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Linq;
using System.Reflection;
using System.Runtime.CompilerServices;
using System.Text;
using System.Threading.Tasks;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Data;
using System.Windows.Input;
using System.Windows.Media;

namespace TestGridViewColumEx
{
    //---------------------------------------------------------
    public delegate void FilterChangedEventHandler();
    public delegate void FilterItemChangedEventHandler(object item);
    //---------------------------------------------------------
    public abstract class BaseFilter : INotifyPropertyChanged
    {
        #region event
        public event FilterChangedEventHandler FilterChanged;
        public event PropertyChangedEventHandler? PropertyChanged;
        public void OnPropertyChanged([CallerMemberName] string propertyName = null)
        {
            PropertyChanged?.Invoke(this, new PropertyChangedEventArgs(propertyName));
            FilterChanged?.Invoke();
        }
        #endregion

        public string Column { get; private set; }
        public bool Locked { get; set; } = false;

        public BaseFilter(string column) => Column = column;

        public virtual bool Check(object value) { return false; }
        public virtual bool IsFilter => false;

        public abstract void Clear();

        private PropertyInfo? _PropertyInfo = null;
        protected PropertyInfo? PropertyInfo(object value)
        {
            if (_PropertyInfo == null)
            {
                _PropertyInfo = value.GetType().GetProperty(Column);
            }
            return _PropertyInfo;
        }
    }
    //---------------------------------------------------------
    public class StringFilter : BaseFilter
    {
        public StringFilter(string column) : base(column) { }

        #region Value
        private object _Value = null;
        public object Value
        {
            get => _Value;
            set { _Value = value; OnPropertyChanged(); }
        }
        #endregion

        public override string ToString()
        {
            return $"STRING FILTER {Column} = \"{Value}\"";
        }

        public override bool IsFilter => !string.IsNullOrEmpty(Value?.ToString());
        public override void Clear() => Value = null;

        public override bool Check(object value)
        {
            var v = PropertyInfo(value)?.GetValue(value);
            return v?.ToString().ToLower().Contains(Value.ToString().ToLower()) ?? false;
        }
    }
    //---------------------------------------------------------
    public class IntegerFilter : StringFilter
    {
        private bool _exact = false;
        public IntegerFilter(string column, bool exact = false) : base(column) => _exact = exact;

        public override bool Check(object value)
        {
            if (!_exact) return base.Check(value);
            else
            {
                var v = PropertyInfo(value)?.GetValue(value);
                return v?.ToString() == Value.ToString().ToLower();
            }
        }
        public override string ToString()
        {
            return $"INTEGER FILTER {Column} = \"{Value}\"";
        }
    }
    //---------------------------------------------------------
    public class TwoIntegerFilter : BaseFilter
    {
        public TwoIntegerFilter(string column) : base(column) { }

        #region ValueFrom
        private string _ValueFrom = null;
        public string ValueFrom
        {
            get => _ValueFrom;
            set { _ValueFrom = value; OnPropertyChanged(); }
        }
        #endregion

        #region ValueTo
        private string _ValueTo = null;
        public string ValueTo
        {
            get => _ValueTo;
            set { _ValueTo = value; OnPropertyChanged(); }
        }
        #endregion

        int from;
        int to;
        public override bool IsFilter
        {
            get
            {
                if (!string.IsNullOrEmpty(ValueFrom) && !string.IsNullOrEmpty(ValueTo))
                {
                    try
                    {
                        from = int.Parse(ValueFrom);
                        to = int.Parse(ValueTo);
                        return from <= to;
                    }
                    catch { }
                }
                return false;
            }
        }

        public override bool Check(object value)
        {
            if (IsFilter)
            {
                var v = (int)PropertyInfo(value)?.GetValue(value);
                return v >= from && v <= to;
            }
            return false;
        }

        public override void Clear()
        {
            ValueFrom = null;
            ValueTo = null;
        }

        public override string ToString()
        {
            return $"TWO INTEGER FILTER {Column} = \"{ValueFrom}\" or \"{ValueTo}\"";
        }
    }
    //---------------------------------------------------------
    public class TwoDateFilter : BaseFilter
    {
        public TwoDateFilter(string column) : base(column) { }

        #region DateFrom
        private DateTime? _DateFrom = null;
        public DateTime? DateFrom
        {
            get => _DateFrom;
            set { _DateFrom = value; OnPropertyChanged(); }
        }
        #endregion

        #region DateTo
        private DateTime? _DateTo = null;
        public DateTime? DateTo
        {
            get => _DateTo;
            set
            {
                if (value != null)
                {
                    var v = (DateTime)value;
                    value = new DateTime(v.Year, v.Month, v.Day, 23, 59, 59);
                }
                _DateTo = value;
                OnPropertyChanged();
            }
        }
        #endregion

        public override bool IsFilter => DateFrom != null && DateTo != null && DateFrom <= DateTo;

        public override bool Check(object value)
        {
            var v = (DateTime)PropertyInfo(value)?.GetValue(value);
            return v >= DateFrom && v <= DateTo;
        }

        public override void Clear()
        {
            DateFrom = null;
            DateTo = null;
        }

        public override string ToString()
        {
            return $"TWO Date FILTER {Column} = \"{DateFrom}\" or \"{DateTo}\"";
        }
    }
    //---------------------------------------------------------
    public class BoolFilter : BaseFilter
    {
        public BoolFilter(string column) : base(column) { }

        #region IsNone
        private bool _IsNone = true;
        public bool IsNone
        {
            get => _IsNone;
            set { _IsNone = value; OnPropertyChanged(); }
        }
        #endregion

        #region IsTrue
        private bool _IsTrue = false;
        public bool IsTrue
        {
            get => _IsTrue;
            set { _IsTrue = value; OnPropertyChanged(); }
        }
        #endregion

        #region IsFalse
        private bool _IsFalse = false;
        public bool IsFalse
        {
            get => _IsFalse;
            set { _IsFalse = value; OnPropertyChanged(); }
        }
        #endregion

        public override string ToString()
        {
            return $"BOOL FILTER {Column} = \"{IsTrue}\"";
        }

        public override bool IsFilter => IsNone == false;
        public override void Clear()
        {
            IsNone = true;
        }

        public override bool Check(object value)
        {
            var v = (bool)PropertyInfo(value)?.GetValue(value);
            return ((IsTrue && v) || (IsFalse && !v));
        }
    }
    //---------------------------------------------------------
    public class StringDateFilter : StringFilter
    {
        private FilterEditor _format = FilterEditor.None;
        public StringDateFilter(string column, FilterEditor format) : base(column) => _format = format;

        public override string ToString()
        {
            return $"STRING DATE FILTER {Column} = \"{Value}\" format = \"{_format.ToString()}\"";
        }
    }
    //---------------------------------------------------------
    public class SpecificDateFilter : StringFilter
    {
        public SpecificDateFilter(string column) : base(column) { }

        public override string ToString()
        {
            return $"DATE FILTER {Column} = \"{Value}\"";
        }

        public override bool Check(object value)
        {
            DateTime v = (DateTime)PropertyInfo(value)?.GetValue(value);
            DateTime x = (DateTime)Value;
            return v.Year == x.Year && v.Month == x.Month && v.Day == x.Day;
        }
    }



    //---------------------------------------------------------
    #region EnumFilter classes
    public abstract class BaseEnumFilter : BaseFilter
    {
        public EnumCheckerList Enums { get; } = new EnumCheckerList();

        public override bool IsFilter => Enums.IsFilter;

        public BaseEnumFilter(string column, Array values,
            string allText = "All", string noneText = "None",
            bool all = false, bool none = false) : base(column)
        {
            if (all) Enums.Add(new EnumCheckerAll { Name = allText });
            foreach (var item in values)
            {
                Enums.Add(new EnumChecker { Name = GetElementString(item), Value = item });
            }
            if (none) Enums.Add(new EnumCheckerNone { Name = noneText });
            Enums.UpdateAllNone();
            Enums.FilterChanged += delegate { OnPropertyChanged("Enum"); };
        }

        protected abstract string GetElementString(object source);

        public override bool Check(object value)
        {
            var v = PropertyInfo(value)?.GetValue(value);
            return Enums[v]?.IsChecked ?? false;
        }

        public override void Clear() => Enums.ClearFilter();
    }
    public class EnumFilter : BaseEnumFilter
    {
        public EnumFilter(string column, Array array,
            string allText = "All", string noneText = "None",
            bool all = false, bool none = false) :
                base(column, array, allText, noneText, all, none)
        {
        }
        protected override string GetElementString(object source) => source?.ToString() ?? "";
    }
    // only for specify differents style editors
    public class StringEnumFilter : EnumFilter
    {
        public StringEnumFilter(string column, Array array,
            string allText = "All", string noneText = "None",
            bool all = false, bool none = false) :
                base(column, array, allText, noneText, all, none)
        { }
    }
    public class IntEnumFilter : EnumFilter
    {
        public IntEnumFilter(string column, Array array,
            string allText = "All", string noneText = "None",
            bool all = false, bool none = false) :
                base(column, array, allText, noneText, all, none)
        { }
    }
    public class DateEnumFilter : EnumFilter
    {
        private FilterEditor _format = FilterEditor.DMY;
        public DateEnumFilter(string column, Array array, FilterEditor format,
            string allText = "All", string noneText = "None",
            bool all = false, bool none = false) :
                base(column, array, allText, noneText, all, none)
        { _format = format; }

        protected override string GetElementString(object source)
        {
            DateTime dt = (DateTime)source;
            var f = _format switch
            {
                FilterEditor.DMY => "dd/MM/yyyy",
                FilterEditor.MDY => "MM/dd/yyyy",
                FilterEditor.YDM => "yyyy/dd/MM",
                FilterEditor.YMD => "yyyy/MM/dd",
            };
            return dt.ToString(f);
        }

        public override bool Check(object value)
        {
            var v = (DateTime)PropertyInfo(value)?.GetValue(value);
            DateTime dt = new DateTime(v.Year, v.Month, v.Day, 0, 0, 0);
            return Enums[dt]?.IsChecked ?? false;
        }
    }
    public class TimeEnumFilter : EnumFilter
    {
        public TimeEnumFilter(string column, Array array,
            string allText = "All", string noneText = "None",
            bool all = false, bool none = false) :
                base(column, array, allText, noneText, all, none)
        { }
    }
    // only for enum
    public class EnumEnumFilter : EnumFilter
    {
        TypeConverter tc = null;
        public EnumEnumFilter(string column, Type type,
            string allText = "All", string noneText = "None",
            bool all = false, bool none = false) :
                base(column, Enum.GetValues(type), allText, noneText, all, none)
        {
        }
        protected override string GetElementString(object source)
        {
            if (tc == null) tc = TypeDescriptor.GetConverter(source.GetType());
            return tc.ConvertToString(source) ?? "";
        }
    }
    //---------------------------------------------------------
    #region class for EnumFilter
    public class EnumCheckerList : List<EnumCheckerBase>
    {
        public event FilterChangedEventHandler FilterChanged;
        private bool disableEvent = false;
        private void OnFilterChanged()
        {
            if (!disableEvent) FilterChanged?.Invoke();
        }

        public EnumCheckerBase this[object o] =>
            this.Where(x => x is EnumChecker && x.Value.Equals(o)).FirstOrDefault();
        public bool IsFilter =>
            this.Where(x => x is EnumChecker && !x.IsChecked).Count() != 0;

        public new void Add(EnumCheckerBase e)
        {
            base.Add(e);
            e.FilterItemChanged += s =>
            {
                if (disableEvent) return;
                disableEvent = true;
                if (s is EnumCheckerAll)
                {
                    this.Where(x => x is EnumChecker).ToList().ForEach(x => x.IsChecked = true);
                }
                else if (s is EnumCheckerNone)
                {
                    this.Where(x => x is EnumChecker).ToList().ForEach(x => x.IsChecked = false);
                }
                disableEvent = false;
                OnFilterChanged();
                UpdateAllNone();
            };
        }
        public void UpdateAllNone()
        {
            disableEvent = true;
            EnumCheckerBase a = this.Where(x => x is EnumCheckerAll).FirstOrDefault();
            EnumCheckerBase n = this.Where(x => x is EnumCheckerNone).FirstOrDefault();
            bool allChecked = this.Where(x => x is EnumChecker && !x.IsChecked).Count() == 0;
            bool noneChecked = this.Where(x => x is EnumChecker && x.IsChecked).Count() == 0;
            if (a != null)
            {
                a.IsChecked = allChecked;
                a.IsEnabled = !allChecked;
            }
            if (n != null)
            {
                n.IsChecked = noneChecked;
                n.IsEnabled = !noneChecked;
            }
            disableEvent = false;
        }

        public void ClearFilter()
        {
            disableEvent = true;
            this.Where(x => x is EnumChecker).ToList().ForEach(x => x.IsChecked = true);
            disableEvent = false;
            OnFilterChanged();
            UpdateAllNone();
        }
    }

    public abstract class EnumCheckerBase : INotifyPropertyChanged
    {
        public event FilterItemChangedEventHandler FilterItemChanged;
        public event PropertyChangedEventHandler? PropertyChanged;

        #region IsChecked
        private bool _IsChecked = true;
        public bool IsChecked
        {
            get => _IsChecked;
            set
            {
                _IsChecked = value;
                FilterItemChanged?.Invoke(this);
                PropertyChanged?.Invoke(this, new PropertyChangedEventArgs("IsChecked"));
            }
        }
        #endregion

        public string Name { get; set; }
        public object Value { get; set; }

        #region IsEnabled
        private bool _IsEnabled = true;
        public bool IsEnabled
        {
            get => _IsEnabled;
            set
            {
                _IsEnabled = value;
                PropertyChanged?.Invoke(this, new PropertyChangedEventArgs("IsEnabled"));
            }
        }
        #endregion
    }
    public class EnumCheckerAll : EnumCheckerBase { }
    public class EnumCheckerNone : EnumCheckerBase { }
    public class EnumChecker : EnumCheckerBase { }
    #endregion
    //---------------------------------------------------------
    #endregion
    //---------------------------------------------------------
}
